==> llm-docs-0.1.4.tgz: package/docs/performance/caching.md <==
---
id: performance-caching
---
# Caching and Performance Guide

<!-- title: Caching and Performance Guide -->
<!-- description: Practical guide for caching and performance optimization in the LLM Platform Platform -->
<!-- lastUpdated: 2024-03-19 -->

## Overview

This guide covers practical caching and performance optimization for the LLM Platform Platform, including Redis caching, memory optimization, and performance monitoring.

## Redis Caching

### Cache Service

```typescript
// src/cache/cache.service.ts
@Injectable()
export class CacheService {
  private readonly redis: Redis;
  private readonly logger = new Logger(CacheService.name);
  
  constructor(
    @Inject(CACHE_MANAGER) private cacheManager: Cache,
    private configService: ConfigService
  ) {
    this.redis = new Redis({
      host: this.configService.get('REDIS_HOST'),
      port: this.configService.get('REDIS_PORT'),
      password: this.configService.get('REDIS_PASSWORD'),
      tls: this.configService.get('REDIS_TLS') === 'true' ? {} : undefined,
      retryStrategy: (times: number) => {
        const delay = Math.min(times * 50, 2000);
        return delay;
      }
    });
    
    this.redis.on('error', (error) => {
      this.logger.error('Redis error:', error);
    });
  }
  
  async get<T>(key: string): Promise<T | null> {
    try {
      const value = await this.redis.get(key);
      return value ? JSON.parse(value) : null;
    } catch (error) {
      this.logger.error(`Cache get error for key ${key}:`, error);
      return null;
    }
  }
  
  async set(
    key: string,
    value: any,
    ttl?: number
  ): Promise<void> {
    try {
      const serialized = JSON.stringify(value);
      if (ttl) {
        await this.redis.set(key, serialized, 'EX', ttl);
      } else {
        await this.redis.set(key, serialized);
      }
    } catch (error) {
      this.logger.error(`Cache set error for key ${key}:`, error);
    }
  }
  
  async del(key: string): Promise<void> {
    try {
      await this.redis.del(key);
    } catch (error) {
      this.logger.error(`Cache delete error for key ${key}:`, error);
    }
  }
  
  async mget<T>(keys: string[]): Promise<(T | null)[]> {
    try {
      const values = await this.redis.mget(keys);
      return values.map(value => 
        value ? JSON.parse(value) : null
      );
    } catch (error) {
      this.logger.error('Cache mget error:', error);
      return keys.map(() => null);
    }
  }
  
  async mset(
    entries: Array<{ key: string; value: any; ttl?: number }>
  ): Promise<void> {
    try {
      const pipeline = this.redis.pipeline();
      
      entries.forEach(({ key, value, ttl }) => {
        const serialized = JSON.stringify(value);
        if (ttl) {
          pipeline.set(key, serialized, 'EX', ttl);
        } else {
          pipeline.set(key, serialized);
        }
      });
      
      await pipeline.exec();
    } catch (error) {
      this.logger.error('Cache mset error:', error);
    }
  }
  
  async scan(
    pattern: string,
    cursor = '0',
    count = 100
  ): Promise<{ cursor: string; keys: string[] }> {
    try {
      const [nextCursor, keys] = await this.redis.scan(
        cursor,
        'MATCH',
        pattern,
        'COUNT',
        count
      );
      
      return {
        cursor: nextCursor,
        keys
      };
    } catch (error) {
      this.logger.error('Cache scan error:', error);
      return { cursor: '0', keys: [] };
    }
  }
}
```

### Cache Interceptor

```typescript
// src/cache/cache.interceptor.ts
@Injectable()
export class CacheInterceptor implements NestInterceptor {
  constructor(
    private cacheService: CacheService,
    private configService: ConfigService
  ) {}
  
  async intercept(
    context: ExecutionContext,
    next: CallHandler
  ): Promise<Observable<any>> {
    const request = context.switchToHttp().getRequest();
    const cacheKey = this.generateCacheKey(request);
    
    // Check cache
    const cached = await this.cacheService.get(cacheKey);
    if (cached) {
      return of(cached);
    }
    
    // Execute request
    return next.handle().pipe(
      tap(async (response) => {
        // Cache response
        const ttl = this.getCacheTTL(request);
        await this.cacheService.set(cacheKey, response, ttl);
      })
    );
  }
  
  private generateCacheKey(request: any): string {
    const { method, url, query, body } = request;
    const key = `${method}:${url}:${JSON.stringify(query)}:${JSON.stringify(body)}`;
    return `cache:${createHash('md5').update(key).digest('hex')}`;
  }
  
  private getCacheTTL(request: any): number {
    // Get TTL from decorator or config
    const ttl = Reflect.getMetadata(
      'cache-ttl',
      request.route?.handler
    ) || this.configService.get('CACHE_TTL');
    
    return parseInt(ttl, 10);
  }
}

// src/decorators/cache.decorator.ts
export const Cache = (ttl?: number) => {
  return (target: any, key: string, descriptor: PropertyDescriptor) => {
    Reflect.defineMetadata('cache-ttl', ttl, descriptor.value);
    return descriptor;
  };
};
```

## Memory Optimization

### Memory Service

```typescript
// src/performance/memory.service.ts
@Injectable()
export class MemoryService {
  private readonly logger = new Logger(MemoryService.name);
  private readonly heapUsedThreshold: number;
  private readonly heapTotalThreshold: number;
  
  constructor(
    private configService: ConfigService
  ) {
    this.heapUsedThreshold = this.configService.get('HEAP_USED_THRESHOLD');
    this.heapTotalThreshold = this.configService.get('HEAP_TOTAL_THRESHOLD');
  }
  
  monitorMemory(): void {
    setInterval(() => {
      const memory = process.memoryUsage();
      
      // Check heap usage
      const heapUsedPercent = (memory.heapUsed / memory.heapTotal) * 100;
      if (heapUsedPercent > this.heapUsedThreshold) {
        this.logger.warn(
          `High heap usage: ${heapUsedPercent.toFixed(2)}%`
        );
      }
      
      // Check total memory
      const totalPercent = (memory.heapTotal / 1024 / 1024 / 1024) * 100;
      if (totalPercent > this.heapTotalThreshold) {
        this.logger.warn(
          `High total memory: ${totalPercent.toFixed(2)}%`
        );
      }
      
      // Log memory stats
      this.logger.debug({
        heapUsed: `${(memory.heapUsed / 1024 / 1024).toFixed(2)} MB`,
        heapTotal: `${(memory.heapTotal / 1024 / 1024).toFixed(2)} MB`,
        external: `${(memory.external / 1024 / 1024).toFixed(2)} MB`,
        rss: `${(memory.rss / 1024 / 1024).toFixed(2)} MB`
      });
    }, 60000); // Check every minute
  }
  
  async forceGC(): Promise<void> {
    if (global.gc) {
      try {
        global.gc();
        this.logger.log('Manual garbage collection completed');
      } catch (error) {
        this.logger.error('Manual garbage collection failed:', error);
      }
    } else {
      this.logger.warn(
        'Manual garbage collection not available. ' +
        'Run with --expose-gc flag'
      );
    }
  }
  
  getMemoryStats(): any {
    const memory = process.memoryUsage();
    return {
      heapUsed: `${(memory.heapUsed / 1024 / 1024).toFixed(2)} MB`,
      heapTotal: `${(memory.heapTotal / 1024 / 1024).toFixed(2)} MB`,
      external: `${(memory.external / 1024 / 1024).toFixed(2)} MB`,
      rss: `${(memory.rss / 1024 / 1024).toFixed(2)} MB`,
      heapUsedPercent: `${((memory.heapUsed / memory.heapTotal) * 100).toFixed(2)}%`
    };
  }
}
```

### Memory Leak Detection

```typescript
// src/performance/leak-detector.service.ts
@Injectable()
export class LeakDetectorService {
  private readonly logger = new Logger(LeakDetectorService.name);
  private readonly snapshots: Map<string, any> = new Map();
  
  takeSnapshot(name: string): void {
    const heap = v8.getHeapSnapshot();
    this.snapshots.set(name, heap);
    
    this.logger.log(`Heap snapshot taken: ${name}`);
  }
  
  compareSnapshots(
    before: string,
    after: string
  ): any {
    const beforeSnapshot = this.snapshots.get(before);
    const afterSnapshot = this.snapshots.get(after);
    
    if (!beforeSnapshot || !afterSnapshot) {
      throw new Error('Snapshots not found');
    }
    
    const comparison = this.analyzeHeapDiff(
      beforeSnapshot,
      afterSnapshot
    );
    
    this.logger.log('Heap comparison:', comparison);
    return comparison;
  }
  
  private analyzeHeapDiff(
    before: any,
    after: any
  ): any {
    const diff: any = {
      totalSize: after.totalSize - before.totalSize,
      totalObjects: after.totalObjects - before.totalObjects,
      leaks: []
    };
    
    // Analyze object growth
    const beforeObjects = this.getObjectCounts(before);
    const afterObjects = this.getObjectCounts(after);
    
    for (const [type, count] of afterObjects) {
      const beforeCount = beforeObjects.get(type) || 0;
      const growth = count - beforeCount;
      
      if (growth > 0) {
        diff.leaks.push({
          type,
          growth,
          percentage: (growth / beforeCount) * 100
        });
      }
    }
    
    return diff;
  }
  
  private getObjectCounts(snapshot: any): Map<string, number> {
    const counts = new Map<string, number>();
    
    // Count objects by type
    snapshot.nodes.forEach((node: any) => {
      const type = node.type;
      counts.set(type, (counts.get(type) || 0) + 1);
    });
    
    return counts;
  }
}
```

## Performance Monitoring

### Performance Service

```typescript
// src/performance/performance.service.ts
@Injectable()
export class PerformanceService {
  private readonly logger = new Logger(PerformanceService.name);
  private readonly metrics: Map<string, number[]> = new Map();
  
  startTimer(name: string): () => void {
    const start = process.hrtime();
    
    return () => {
      const [seconds, nanoseconds] = process.hrtime(start);
      const duration = seconds * 1000 + nanoseconds / 1000000;
      
      this.recordMetric(name, duration);
    };
  }
  
  recordMetric(name: string, value: number): void {
    if (!this.metrics.has(name)) {
      this.metrics.set(name, []);
    }
    
    this.metrics.get(name)!.push(value);
    
    // Keep last 1000 values
    if (this.metrics.get(name)!.length > 1000) {
      this.metrics.get(name)!.shift();
    }
  }
  
  getMetrics(): any {
    const result: any = {};
    
    for (const [name, values] of this.metrics) {
      result[name] = {
        count: values.length,
        min: Math.min(...values),
        max: Math.max(...values),
        avg: values.reduce((a, b) => a + b, 0) / values.length,
        p95: this.percentile(values, 95),
        p99: this.percentile(values, 99)
      };
    }
    
    return result;
  }
  
  private percentile(values: number[], p: number): number {
    const sorted = [...values].sort((a, b) => a - b);
    const index = Math.ceil((p / 100) * sorted.length) - 1;
    return sorted[index];
  }
}

// src/interceptors/performance.interceptor.ts
@Injectable()
export class PerformanceInterceptor implements NestInterceptor {
  constructor(
    private performanceService: PerformanceService,
    private logger: Logger
  ) {}
  
  intercept(
    context: ExecutionContext,
    next: CallHandler
  ): Observable<any> {
    const request = context.switchToHttp().getRequest();
    const { method, url } = request;
    const metricName = `${method}:${url}`;
    
    const stopTimer = this.performanceService.startTimer(metricName);
    
    return next.handle().pipe(
      tap({
        next: () => {
          stopTimer();
        },
        error: (error) => {
          stopTimer();
          this.logger.error(
            `Request failed: ${method} ${url}`,
            error
          );
        }
      })
    );
  }
}
```

## Best Practices

1. **Caching Strategy**
   - Use appropriate TTLs
   - Cache invalidation
   - Cache warming
   - Cache monitoring

2. **Memory Management**
   - Monitor heap usage
   - Detect memory leaks
   - Optimize data structures
   - Regular garbage collection

3. **Performance Optimization**
   - Measure performance
   - Profile bottlenecks
   - Optimize critical paths
   - Monitor metrics

4. **Resource Management**
   - Connection pooling
   - Resource limits
   - Error handling
   - Circuit breakers

## Related Documentation

- [Architecture Guide](../architecture/README)
- [Development Guide](../development/README)
- [Deployment Guide](../deployment/README)
- [Database Guide](../database/README)
- [Operations Guide](../operations/README)

==> llm-docs-1.0.1.tgz: package/docs/performance/caching.md <==
---
id: performance-caching
---
# Caching and Performance Guide

<!-- title: Caching and Performance Guide -->
<!-- description: Practical guide for caching and performance optimization in the Bluefly Platform -->
<!-- lastUpdated: 2024-03-19 -->

## Overview

This guide covers practical caching and performance optimization for the Bluefly Platform, including Redis caching, memory optimization, and performance monitoring.

## Redis Caching

### Cache Service

```typescript
// src/cache/cache.service.ts
@Injectable()
export class CacheService {
  private readonly redis: Redis;
  private readonly logger = new Logger(CacheService.name);
  
  constructor(
    @Inject(CACHE_MANAGER) private cacheManager: Cache,
    private configService: ConfigService
  ) {
    this.redis = new Redis({
      host: this.configService.get('REDIS_HOST'),
      port: this.configService.get('REDIS_PORT'),
      password: this.configService.get('REDIS_PASSWORD'),
      tls: this.configService.get('REDIS_TLS') === 'true' ? {} : undefined,
      retryStrategy: (times: number) => {
        const delay = Math.min(times * 50, 2000);
        return delay;
      }
    });
    
    this.redis.on('error', (error) => {
      this.logger.error('Redis error:', error);
    });
  }
  
  async get<T>(key: string): Promise<T | null> {
    try {
      const value = await this.redis.get(key);
      return value ? JSON.parse(value) : null;
    } catch (error) {
      this.logger.error(`Cache get error for key ${key}:`, error);
      return null;
    }
  }
  
  async set(
    key: string,
    value: any,
    ttl?: number
  ): Promise<void> {
    try {
      const serialized = JSON.stringify(value);
      if (ttl) {
        await this.redis.set(key, serialized, 'EX', ttl);
      } else {
        await this.redis.set(key, serialized);
      }
    } catch (error) {
      this.logger.error(`Cache set error for key ${key}:`, error);
    }
  }
  
  async del(key: string): Promise<void> {
    try {
      await this.redis.del(key);
    } catch (error) {
      this.logger.error(`Cache delete error for key ${key}:`, error);
    }
  }
  
  async mget<T>(keys: string[]): Promise<(T | null)[]> {
    try {
      const values = await this.redis.mget(keys);
      return values.map(value => 
        value ? JSON.parse(value) : null
      );
    } catch (error) {
      this.logger.error('Cache mget error:', error);
      return keys.map(() => null);
    }
  }
  
  async mset(
    entries: Array<{ key: string; value: any; ttl?: number }>
  ): Promise<void> {
    try {
      const pipeline = this.redis.pipeline();
      
      entries.forEach(({ key, value, ttl }) => {
        const serialized = JSON.stringify(value);
        if (ttl) {
          pipeline.set(key, serialized, 'EX', ttl);
        } else {
          pipeline.set(key, serialized);
        }
      });
      
      await pipeline.exec();
    } catch (error) {
      this.logger.error('Cache mset error:', error);
    }
  }
  
  async scan(
    pattern: string,
    cursor = '0',
    count = 100
  ): Promise<{ cursor: string; keys: string[] }> {
    try {
      const [nextCursor, keys] = await this.redis.scan(
        cursor,
        'MATCH',
        pattern,
        'COUNT',
        count
      );
      
      return {
        cursor: nextCursor,
        keys
      };
    } catch (error) {
      this.logger.error('Cache scan error:', error);
      return { cursor: '0', keys: [] };
    }
  }
}
```

### Cache Interceptor

```typescript
// src/cache/cache.interceptor.ts
@Injectable()
export class CacheInterceptor implements NestInterceptor {
  constructor(
    private cacheService: CacheService,
    private configService: ConfigService
  ) {}
  
  async intercept(
    context: ExecutionContext,
    next: CallHandler
  ): Promise<Observable<any>> {
    const request = context.switchToHttp().getRequest();
    const cacheKey = this.generateCacheKey(request);
    
    // Check cache
    const cached = await this.cacheService.get(cacheKey);
    if (cached) {
      return of(cached);
    }
    
    // Execute request
    return next.handle().pipe(
      tap(async (response) => {
        // Cache response
        const ttl = this.getCacheTTL(request);
        await this.cacheService.set(cacheKey, response, ttl);
      })
    );
  }
  
  private generateCacheKey(request: any): string {
    const { method, url, query, body } = request;
    const key = `${method}:${url}:${JSON.stringify(query)}:${JSON.stringify(body)}`;
    return `cache:${createHash('md5').update(key).digest('hex')}`;
  }
  
  private getCacheTTL(request: any): number {
    // Get TTL from decorator or config
    const ttl = Reflect.getMetadata(
      'cache-ttl',
      request.route?.handler
    ) || this.configService.get('CACHE_TTL');
    
    return parseInt(ttl, 10);
  }
}

// src/decorators/cache.decorator.ts
export const Cache = (ttl?: number) => {
  return (target: any, key: string, descriptor: PropertyDescriptor) => {
    Reflect.defineMetadata('cache-ttl', ttl, descriptor.value);
    return descriptor;
  };
};
```

## Memory Optimization

### Memory Service

```typescript
// src/performance/memory.service.ts
@Injectable()
export class MemoryService {
  private readonly logger = new Logger(MemoryService.name);
  private readonly heapUsedThreshold: number;
  private readonly heapTotalThreshold: number;
  
  constructor(
    private configService: ConfigService
  ) {
    this.heapUsedThreshold = this.configService.get('HEAP_USED_THRESHOLD');
    this.heapTotalThreshold = this.configService.get('HEAP_TOTAL_THRESHOLD');
  }
  
  monitorMemory(): void {
    setInterval(() => {
      const memory = process.memoryUsage();
      
      // Check heap usage
      const heapUsedPercent = (memory.heapUsed / memory.heapTotal) * 100;
      if (heapUsedPercent > this.heapUsedThreshold) {
        this.logger.warn(
          `High heap usage: ${heapUsedPercent.toFixed(2)}%`
        );
      }
      
      // Check total memory
      const totalPercent = (memory.heapTotal / 1024 / 1024 / 1024) * 100;
      if (totalPercent > this.heapTotalThreshold) {
        this.logger.warn(
          `High total memory: ${totalPercent.toFixed(2)}%`
        );
      }
      
      // Log memory stats
      this.logger.debug({
        heapUsed: `${(memory.heapUsed / 1024 / 1024).toFixed(2)} MB`,
        heapTotal: `${(memory.heapTotal / 1024 / 1024).toFixed(2)} MB`,
        external: `${(memory.external / 1024 / 1024).toFixed(2)} MB`,
        rss: `${(memory.rss / 1024 / 1024).toFixed(2)} MB`
      });
    }, 60000); // Check every minute
  }
  
  async forceGC(): Promise<void> {
    if (global.gc) {
      try {
        global.gc();
        this.logger.log('Manual garbage collection completed');
      } catch (error) {
        this.logger.error('Manual garbage collection failed:', error);
      }
    } else {
      this.logger.warn(
        'Manual garbage collection not available. ' +
        'Run with --expose-gc flag'
      );
    }
  }
  
  getMemoryStats(): any {
    const memory = process.memoryUsage();
    return {
      heapUsed: `${(memory.heapUsed / 1024 / 1024).toFixed(2)} MB`,
      heapTotal: `${(memory.heapTotal / 1024 / 1024).toFixed(2)} MB`,
      external: `${(memory.external / 1024 / 1024).toFixed(2)} MB`,
      rss: `${(memory.rss / 1024 / 1024).toFixed(2)} MB`,
      heapUsedPercent: `${((memory.heapUsed / memory.heapTotal) * 100).toFixed(2)}%`
    };
  }
}
```

### Memory Leak Detection

```typescript
// src/performance/leak-detector.service.ts
@Injectable()
export class LeakDetectorService {
  private readonly logger = new Logger(LeakDetectorService.name);
  private readonly snapshots: Map<string, any> = new Map();
  
  takeSnapshot(name: string): void {
    const heap = v8.getHeapSnapshot();
    this.snapshots.set(name, heap);
    
    this.logger.log(`Heap snapshot taken: ${name}`);
  }
  
  compareSnapshots(
    before: string,
    after: string
  ): any {
    const beforeSnapshot = this.snapshots.get(before);
    const afterSnapshot = this.snapshots.get(after);
    
    if (!beforeSnapshot || !afterSnapshot) {
      throw new Error('Snapshots not found');
    }
    
    const comparison = this.analyzeHeapDiff(
      beforeSnapshot,
      afterSnapshot
    );
    
    this.logger.log('Heap comparison:', comparison);
    return comparison;
  }
  
  private analyzeHeapDiff(
    before: any,
    after: any
  ): any {
    const diff: any = {
      totalSize: after.totalSize - before.totalSize,
      totalObjects: after.totalObjects - before.totalObjects,
      leaks: []
    };
    
    // Analyze object growth
    const beforeObjects = this.getObjectCounts(before);
    const afterObjects = this.getObjectCounts(after);
    
    for (const [type, count] of afterObjects) {
      const beforeCount = beforeObjects.get(type) || 0;
      const growth = count - beforeCount;
      
      if (growth > 0) {
        diff.leaks.push({
          type,
          growth,
          percentage: (growth / beforeCount) * 100
        });
      }
    }
    
    return diff;
  }
  
  private getObjectCounts(snapshot: any): Map<string, number> {
    const counts = new Map<string, number>();
    
    // Count objects by type
    snapshot.nodes.forEach((node: any) => {
      const type = node.type;
      counts.set(type, (counts.get(type) || 0) + 1);
    });
    
    return counts;
  }
}
```

## Performance Monitoring

### Performance Service

```typescript
// src/performance/performance.service.ts
@Injectable()
export class PerformanceService {
  private readonly logger = new Logger(PerformanceService.name);
  private readonly metrics: Map<string, number[]> = new Map();
  
  startTimer(name: string): () => void {
    const start = process.hrtime();
    
    return () => {
      const [seconds, nanoseconds] = process.hrtime(start);
      const duration = seconds * 1000 + nanoseconds / 1000000;
      
      this.recordMetric(name, duration);
    };
  }
  
  recordMetric(name: string, value: number): void {
    if (!this.metrics.has(name)) {
      this.metrics.set(name, []);
    }
    
    this.metrics.get(name)!.push(value);
    
    // Keep last 1000 values
    if (this.metrics.get(name)!.length > 1000) {
      this.metrics.get(name)!.shift();
    }
  }
  
  getMetrics(): any {
    const result: any = {};
    
    for (const [name, values] of this.metrics) {
      result[name] = {
        count: values.length,
        min: Math.min(...values),
        max: Math.max(...values),
        avg: values.reduce((a, b) => a + b, 0) / values.length,
        p95: this.percentile(values, 95),
        p99: this.percentile(values, 99)
      };
    }
    
    return result;
  }
  
  private percentile(values: number[], p: number): number {
    const sorted = [...values].sort((a, b) => a - b);
    const index = Math.ceil((p / 100) * sorted.length) - 1;
    return sorted[index];
  }
}

// src/interceptors/performance.interceptor.ts
@Injectable()
export class PerformanceInterceptor implements NestInterceptor {
  constructor(
    private performanceService: PerformanceService,
    private logger: Logger
  ) {}
  
  intercept(
    context: ExecutionContext,
    next: CallHandler
  ): Observable<any> {
    const request = context.switchToHttp().getRequest();
    const { method, url } = request;
    const metricName = `${method}:${url}`;
    
    const stopTimer = this.performanceService.startTimer(metricName);
    
    return next.handle().pipe(
      tap({
        next: () => {
          stopTimer();
        },
        error: (error) => {
          stopTimer();
          this.logger.error(
            `Request failed: ${method} ${url}`,
            error
          );
        }
      })
    );
  }
}
```

## Best Practices

1. **Caching Strategy**
   - Use appropriate TTLs
   - Cache invalidation
   - Cache warming
   - Cache monitoring

2. **Memory Management**
   - Monitor heap usage
   - Detect memory leaks
   - Optimize data structures
   - Regular garbage collection

3. **Performance Optimization**
   - Measure performance
   - Profile bottlenecks
   - Optimize critical paths
   - Monitor metrics

4. **Resource Management**
   - Connection pooling
   - Resource limits
   - Error handling
   - Circuit breakers

## Related Documentation

- [Architecture Guide](../architecture/README)
- [Development Guide](../development/README)
- [Deployment Guide](../deployment/README)
- [Database Guide](../database/README)
- [Operations Guide](../operations/README)
